"""Defines a class for flexible functional mask generation."""
import os
import os.path as op
import shutil
from tempfile import mkdtemp
from subprocess import check_output
from IPython.parallel import Client
from IPython.parallel.error import TimeoutError

import numpy as np
import nibabel as nib
import matplotlib as mpl
mpl.use("Agg")

from moss.mosaic import Mosaic
import lyman


class MaskFactory(object):
    """Class for the rapid and flexible creation of functional masks.

    This class can make appropriate calls to external (Freesurfer and
    FSL) command-line programs to take ROIs defined in a variety of
    sources and generate binary mask images in native EPI space.

    """
    def __init__(self, subject_list, experiment, roi_name,
                 orig_type, force_serial=False, debug=False):

        # Set up basic info
        self.subject_list = lyman.determine_subjects(subject_list)
        project = lyman.gather_project_info()
        self.experiment = experiment
        self.roi_name = roi_name
        self.orig_type = orig_type
        self.debug = debug
        if debug:
            print "Setting up for %d subjects" % len(subject_list)
            print "Experiment name:", experiment
            print "ROI name:", roi_name

        # Set up directories
        if project["default_exp"] is not None and experiment is None:
            experiment = project["default_exp"]
        self.experiment = experiment
        self.data_dir = project["data_dir"]
        self.anal_dir = project["analysis_dir"]

        # Set up temporary output
        self.temp_dir = mkdtemp()

        # Set the SUBJECTS_DIR variable for Freesurfer
        os.environ["SUBJECTS_DIR"] = self.data_dir

        # Set up parallel execution
        self.parallel = False
        if force_serial:
            self.map = map
        else:
            try:
                rc = Client()
                self.dv = rc[:]
                self.map = self.dv.map_async
                # Push SUBJECTS_DIR to engines
                self.dv.execute("import os")
                self.dv["data_dir"] = self.data_dir
                self.dv.execute("os.environ['SUBJECTS_DIR'] = data_dir")
                self.parallel = True

            except (TimeoutError, IOError):
                self.map = map
        if debug:
            print "Set to run in %s" % (
                "parallel" if self.parallel else "serial")

        # Set up some persistent templates
        self.epi_template = op.join(self.anal_dir, self.experiment,
                                    "%(subj)s",
                                    "preproc/run_1/mean_func.nii.gz")
        self.fov_template = op.join(self.anal_dir, self.experiment,
                                    "%(subj)s",
                                    "preproc/run_1/functional_mask.nii.gz")
        self.reg_template = op.join(self.anal_dir, self.experiment,
                                    "%(subj)s",
                                    "preproc/run_1/func2anat_tkreg.dat")
        self.out_template = op.join(self.data_dir,
                                    "%(subj)s",
                                    "masks/%s.nii.gz" % self.roi_name)
        if debug:
            print "EPI template: %s" % self.epi_template
            print "Reg template: %s" % self.reg_template
            print "Output template: %s" % self.out_template

        # Ensure the output directory will exist
        for subj in self.subject_list:
            mask_dir = op.join(self.data_dir, subj, "masks")
            if not op.exists(mask_dir):
                os.mkdir(mask_dir)

    def __del__(self):

        if self.debug:
            print "Debug mode: not removing output directory:"
            print self.temp_dir
        else:
            shutil.rmtree(self.temp_dir)

    def from_common_label(self, label_template, hemis, proj_args,
                          save_native=False):
        """Reverse normalize possibly bilateral labels to native space."""
        native_label_temp = op.join(self.temp_dir,
                                    "%(hemi)s.%(subj)s_native_label.label")

        # Transform by subject and hemi
        warp_cmds = []
        for subj in self.subject_list:
            for hemi in hemis:
                args = dict(hemi=hemi, subj=subj)
                cmd = ["mri_label2label",
                       "--srcsubject", "fsaverage",
                       "--trgsubject", subj,
                       "--hemi", hemi,
                       "--srclabel", label_template % args,
                       "--trglabel", native_label_temp % args,
                       "--regmethod", "surface"]
                warp_cmds.append(cmd)

        # Execute the transformation
        self.execute(warp_cmds, native_label_temp)

        # Possibly copy the resulting native space label to
        # the subject's label directory
        if save_native:
            save_temp = op.join(self.data_dir, "%(subj)s", "label",
                                "%(hemi)s." + self.roi_name + ".label")
            for subj in self.subject_list:
                for hemi in hemis:
                    args = dict(subj=subj, hemi=hemi)
                    shutil.copyfile(native_label_temp % args, save_temp % args)

        # Carry on with the native label stage
        self.from_native_label(native_label_temp, hemis, proj_args)

    def from_native_label(self, label_template, hemis, proj_args):
        """Given possibly bilateral native labels, make epi masks."""
        indiv_mask_temp = op.join(self.temp_dir,
                                  "%(hemi)s.%(subj)s_mask.nii.gz")
        # Command list for this step
        proj_cmds = []
        for subj in self.subject_list:
            for hemi in hemis:
                args = dict(hemi=hemi, subj=subj)
                cmd = ["mri_label2vol",
                       "--label", label_template % args,
                       "--temp", self.epi_template % args,
                       "--reg", self.reg_template % args,
                       "--hemi", hemi,
                       "--subject", subj,
                       "--o", indiv_mask_temp % args,
                       "--proj"]
                cmd.extend(proj_args)
                proj_cmds.append(cmd)

        # Execute the projection from a surface label
        self.execute(proj_cmds, indiv_mask_temp)

        # Combine the bilateral masks into the final mask
        combine_cmds = []
        for subj in self.subject_list:
            cmd = ["mri_concat"]
            for hemi in hemis:
                args = dict(hemi=hemi, subj=subj)
                cmd.append(indiv_mask_temp % args)
            cmd.extend(["--max",
                        "--o", self.out_template % dict(subj=subj)])
            combine_cmds.append(cmd)

        # Execute the final step
        self.execute(combine_cmds, self.out_template)

    def from_hires_atlas(self, hires_atlas_template, region_ids, erode):
        """Create epi space mask from index volume (e.g. aseg.mgz"""
        hires_mask_template = op.join(self.temp_dir,
                                      "%(subj)s_hires_mask.nii.gz")

        # First run mri_binarize
        bin_cmds = []
        for subj in self.subject_list:
            args = dict(subj=subj)
            cmd_list = ["mri_binarize",
                        "--i", hires_atlas_template % args,
                        "--o", hires_mask_template % args]
            for id in region_ids:
                cmd_list.extend(["--match", str(id)])
            if erode is not None:
                cmd_list.extend(["--erode", str(erode)])
            bin_cmds.append(cmd_list)
        self.execute(bin_cmds, hires_mask_template)

        self.from_hires_mask(hires_mask_template)

    def from_hires_mask(self, hires_mask_template):
        """Create epi space mask from hires mask (binary) volume."""
        xfm_cmds = []
        for subj in self.subject_list:
            args = dict(subj=subj)
            xfm_cmds.append(
                ["mri_vol2vol",
                 "--mov", self.epi_template % args,
                 "--targ", hires_mask_template % args,
                 "--inv",
                 "--o", self.out_template % args,
                 "--reg", self.reg_template % args,
                 "--no-save-reg",
                 "--nearest"])
        self.execute(xfm_cmds, self.out_template)

    def apply_statistical_mask(self, stat_file_temp,
                               thresh=None, n_voxels=None):
        """Create a mask by binarizing an epi-space fixed effects zstat map."""
        if n_voxels is not None:
            return self.take_top_voxels(stat_file_temp, n_voxels)

        bin_cmds = []
        for subj in self.subject_list:
            args = dict(subj=subj)
            cmd = ["fslmaths",
                   stat_file_temp % args,
                   "-thr", thresh,
                   "-bin",
                   "-mul",
                   self.out_template % args,
                   self.out_template % args]
            bin_cmds.append(cmd)

        self.execute(bin_cmds, self.out_template)

    def take_top_voxels(self, stat_file_temp, n_voxels):
        """Create take the top n voxels within a mask based on a stat map."""
        for subj in self.subject_list:
            args = dict(subj=subj)
            mask_file = self.out_template % args
            stat_file = stat_file_temp % args
            img = nib.load(mask_file)
            mask = img.get_data().astype(bool)
            stat = nib.load(stat_file).get_data()
            stat[~mask] = -np.inf
            not_mask = np.argsort(stat.flat).argsort() < (mask.size - n_voxels)
            mask.flat[not_mask] = 0
            new_img = nib.Nifti1Image(mask, img.get_affine(), img.get_header())
            new_img.to_filename(mask_file)

            if self.debug:
                print "Mask file %s sum: %d" % (mask_file, mask.sum())

    def write_png(self):
        """Write a mosiac png showing the masked voxels."""
        slices_temp = op.join(self.data_dir, "%(subj)s/masks",
                              self.roi_name + ".png")

        for subj in self.subject_list:
            args = dict(subj=subj)
            m = Mosaic(self.epi_template % args,
                       self.out_template % args,
                       self.fov_template % args,
                       step=1, show_mask=False,
                       stat_interp="nearest")
            cmap = mpl.colors.ListedColormap(["#C41E3A"])
            m.plot_overlay(cmap, thresh=.5, alpha=.9, colorbar=False)
            m.savefig(slices_temp % args)
            m.close()

    def execute(self, cmd_list, out_temp):
        """Exceute a list of commands and verify output file existence."""
        res = self.map(check_output, cmd_list)
        if self.parallel:
            if self.debug:
                res.wait_interactive()
            else:
                res.wait()
        if self.parallel:
            if not res.successful():
                raise RuntimeError(res.pyerr)



import os
import json
import os.path as op
from nipype.utils.filemanip import fname_presuffix
from nipype.interfaces.base import (BaseInterface, BaseInterfaceInputSpec,
                                    TraitedSpec, InputMultiPath, File, traits)


class SaveParametersInput(BaseInterfaceInputSpec):

    exp_info = traits.Dict()
    in_file = traits.Either(InputMultiPath(File(exists=True)),
                            File(exists=True))


class SaveParametersOutput(TraitedSpec):

    json_file = File(exists=True)


class SaveParameters(BaseInterface):

    input_spec = SaveParametersInput
    output_spec = SaveParametersOutput

    def _run_interface(self, runtime):

        with open("experiment_info.json", "w") as fp:
            json.dump(self.inputs.exp_info, fp, sort_keys=True, indent=2)

        return runtime

    def _list_outputs(self):

        outputs = self._outputs().get()
        outputs["json_file"] = op.abspath("experiment_info.json")
        return outputs


def dump_exp_info(exp_info, timeseries):
    """Dump the exp_info dict into a json file."""
    json_file = op.abspath("experiment_info.json")
    with open(json_file, "w") as fp:
        json.dump(exp_info, fp, sort_keys=True, indent=2)
        return json_file


def add_suffix(fname, suffix):
    """Insert a suffix into a filename before the extension."""
    out_fname = fname_presuffix(fname, suffix="_" + suffix,
                                use_ext=True)
    return out_fname


def nii_to_png(fname, suffix=""):
    """Return a path to write a local png based on an image."""
    out_fname = fname_presuffix(fname, suffix=suffix + ".png",
                                newpath=os.getcwd(),
                                use_ext=False)
    return out_fname



import numpy as np
import nibabel as nib
import nose.tools as nt
import numpy.testing as npt
from nipype.interfaces import fsl

from .. import mixedfx


class TestGroupMerge(object):

    rs = np.random.RandomState(77)

    def test_group_mask(self):

        # Load the common-space mask
        mni_mask = fsl.Info.standard_image("MNI152_T1_2mm_brain_mask.nii.gz")
        mni_img = nib.load(mni_mask)
        mask_data = mni_img.get_data().astype(bool)

        # Create some "variance" data with sparsity
        i, j, k = mask_data.shape
        var_data = self.rs.gamma(3, size=(i, j, k, 10))
        zero_voxels = self.rs.uniform(size=(i, j, k, 10)) < .5
        var_data[zero_voxels] = 0
        var_img = nib.Nifti1Image(var_data, mni_img.get_affine())

        merge = mixedfx.MergeAcrossSubjects()
        out_mask = merge._create_group_mask(var_img)
        out_data = out_mask.get_data().astype(bool)
        nt.assert_is_instance(out_mask, nib.Nifti1Image)

        # Test that the final mask is a subset of both the MNI space mask
        # and of a mask derived from all voxels with nonzero variance
        nt.assert_equal(np.sum(~mask_data & out_data), 0)
        nt.assert_equal(np.sum(~(var_data > 0).all(axis=-1) & out_data), 0)

    def test_group_merge(self):

        img_size = (91, 109, 91)
        in_data = [self.rs.normal(size=img_size) for _ in range(10)]
        in_imgs = [nib.Nifti1Image(d, np.eye(4)) for d in in_data]

        merge = mixedfx.MergeAcrossSubjects()
        out_img = merge._merge_subject_images(in_imgs)
        nt.assert_is_instance(out_img, nib.Nifti1Image)
        nt.assert_equal(out_img.shape, (91, 109, 91, 10))
        out_data = out_img.get_data()
        in_data_reordered = np.asarray(in_data).transpose(1, 2, 3, 0)
        npt.assert_array_equal(in_data_reordered, out_data)



"""Fixed effects model to combine across runs for a single subject."""
import os
import os.path as op
import numpy as np
import nibabel as nib
from moss.mosaic import Mosaic

from nipype import Workflow, Node, IdentityInterface
from nipype.interfaces import fsl
from nipype.interfaces.base import (BaseInterface,
                                    BaseInterfaceInputSpec,
                                    TraitedSpec, File, Directory,
                                    InputMultiPath, OutputMultiPath,
                                    traits)

import lyman
from lyman.tools import SaveParameters, nii_to_png, submit_cmdline


def create_ffx_workflow(name="mni_ffx", space="mni",
                        contrasts=None, exp_info=None):
    """Return a workflow object to execute a fixed-effects mode."""
    if contrasts is None:
        contrasts = []
    if exp_info is None:
        exp_info = lyman.default_experiment_parameters()

    inputnode = Node(IdentityInterface(["copes",
                                        "varcopes",
                                        "masks",
                                        "means",
                                        "dofs",
                                        "ss_files",
                                        "anatomy",
                                        "reg_file",
                                        "timeseries"]),
                     name="inputnode")

    # Fit the fixedfx model for each contrast
    ffxmodel = Node(FFXModel(contrasts=contrasts), "ffxmodel")

    # Calculate the fixed effects Rsquared maps
    ffxsummary = Node(FFXSummary(), "ffxsummary")

    # Plot the fixedfx results
    report = Node(FFXReport(space=space), "report")

    # Save the experiment info
    saveparams = Node(SaveParameters(exp_info=exp_info), "saveparams")

    outputnode = Node(IdentityInterface(["flame_results",
                                         "r2_files",
                                         "tsnr_file",
                                         "summary_report",
                                         "json_file",
                                         "zstat_report"]),
                      "outputs")

    ffx = Workflow(name=name)
    ffx.connect([
        (inputnode, ffxmodel,
            [("copes", "copes"),
             ("varcopes", "varcopes"),
             ("dofs", "dofs"),
             ("masks", "masks"),
             ("reg_file", "reg_file")]),
        (inputnode, ffxsummary,
            [("ss_files", "ss_files"),
             ("means", "means"),
             ("timeseries", "timeseries")]),
        (inputnode, report,
            [("anatomy", "anatomy"),
             ("masks", "masks")]),
        (inputnode, saveparams,
            [("timeseries", "in_file")]),
        (ffxmodel, report,
            [("zstat_files", "zstat_files")]),
        (ffxsummary, report,
            [("r2_files", "r2_files"),
             ("tsnr_file", "tsnr_file")]),
        (ffxmodel, outputnode,
            [("flame_results", "flame_results")]),
        (ffxsummary, outputnode,
            [("r2_files", "r2_files"),
             ("tsnr_file", "tsnr_file")]),
        (report, outputnode,
            [("summary_files", "summary_report"),
             ("zstat_files", "zstat_report")]),
        (saveparams, outputnode,
            [("json_file", "json_file")]),
    ])

    return ffx, inputnode, outputnode


# =========================================================================== #


class FFXModelInput(BaseInterfaceInputSpec):

    contrasts = traits.List()
    copes = InputMultiPath(File(exists=True))
    varcopes = InputMultiPath(File(exists=True))
    dofs = InputMultiPath(File(exists=True))
    masks = InputMultiPath(File(exists=True))
    reg_file = File(exists=True)


class FFXModelOutput(TraitedSpec):

    flame_results = OutputMultiPath(Directory(exists=True))
    zstat_files = OutputMultiPath(File(exists=True))


class FFXModel(BaseInterface):

    input_spec = FFXModelInput
    output_spec = FFXModelOutput

    def _run_interface(self, runtime):

        n_con = len(self.inputs.contrasts)

        # Find the basic geometry of the image
        img = nib.load(self.inputs.copes[0])
        x, y, z = img.shape
        aff, hdr = img.get_affine(), img.get_header()

        # Get lists of files for each contrast
        copes = self._unpack_files(self.inputs.copes, n_con)
        varcopes = self._unpack_files(self.inputs.varcopes, n_con)

        # Make an image with the DOF for each run
        dofs = np.array([np.loadtxt(f) for f in self.inputs.dofs])
        dof_data = np.ones((x, y, z, len(dofs))) * dofs

        # Find the intersection of the masks
        mask_data = [nib.load(f).get_data() for f in self.inputs.masks]
        common_mask = np.all(mask_data, axis=0)
        nib.Nifti1Image(common_mask, aff, hdr).to_filename("mask.nii.gz")

        # Run the flame models
        flame_results = []
        zstat_files = []
        for i, contrast in enumerate(self.inputs.contrasts):

            # Load each run of cope and varcope files into a list
            cs = [nib.load(f).get_data()[..., np.newaxis] for f in copes[i]]
            vs = [nib.load(f).get_data()[..., np.newaxis] for f in varcopes[i]]

            # Find all of the nonzero copes
            # This handles cases where there were no events for some of
            # the runs for the contrast we're currently dealing with
            good_cs = [not np.allclose(d, 0) for d in cs]
            good_vs = [not np.allclose(d, 0) for d in vs]
            good = np.all([good_cs, good_vs], axis=0)

            # Concatenate the cope and varcope data, save only the good frames
            c_data = np.concatenate(cs, axis=-1)[:, :, :, good]
            v_data = np.concatenate(vs, axis=-1)[:, :, :, good]

            # Write out the concatenated copes and varcopes
            nib.Nifti1Image(c_data, aff, hdr).to_filename("cope_4d.nii.gz")
            nib.Nifti1Image(v_data, aff, hdr).to_filename("varcope_4d.nii.gz")

            # Write out a correctly sized design for this contrast
            fsl.L2Model(num_copes=int(good.sum())).run()

            # Mask the DOF data and write it out for this run
            contrast_dof = dof_data[:, :, :, good]
            nib.Nifti1Image(contrast_dof, aff, hdr).to_filename("dof.nii.gz")

            # Build the flamo commandline and run
            flamecmd = ["flameo",
                        "--cope=cope_4d.nii.gz",
                        "--varcope=varcope_4d.nii.gz",
                        "--mask=mask.nii.gz",
                        "--dvc=dof.nii.gz",
                        "--runmode=fe",
                        "--dm=design.mat",
                        "--tc=design.con",
                        "--cs=design.grp",
                        "--ld=" + contrast,
                        "--npo"]
            runtime = submit_cmdline(runtime, flamecmd)

            # Rename the written file and append to the outputs
            for kind in ["cope", "varcope"]:
                os.rename(kind + "_4d.nii.gz",
                          "%s/%s_4d.nii.gz" % (contrast, kind))

            # Put the zstats and mask on the surface
            for hemi in ["lh", "rh"]:
                projcmd = ["mri_vol2surf",
                           "--mov", "%s/zstat1.nii.gz" % contrast,
                           "--reg", self.inputs.reg_file,
                           "--surf-fwhm", "5",
                           "--hemi", hemi,
                           "--projfrac-avg", "0", "1", ".1",
                           "--o", "%s/%s.zstat1.mgz" % (contrast, hemi)]
                submit_cmdline(runtime, projcmd)

                # Mask image
                projcmd = ["mri_vol2surf",
                           "--mov", "%s/mask.nii.gz" % contrast,
                           "--reg", self.inputs.reg_file,
                           "--hemi", hemi,
                           "--projfrac-max", "0", "1", ".1",
                           "--o", "%s/%s.mask.mgz" % (contrast, hemi)]
                submit_cmdline(runtime, projcmd)

            flame_results.append(op.abspath(contrast))
            zstat_files.append(op.abspath("%s/zstat1.nii.gz" % contrast))

        self.flame_results = flame_results
        self.zstat_files = zstat_files

        return runtime

    def _unpack_files(self, files, n_con):
        """Unpack a list of (var)copes into a list of lists by contrast."""
        out = []
        for con in range(n_con):
            out.append([f for f in files if "cope{}_".format(con + 1) in f])
        return out

    def _list_outputs(self):

        outputs = self._outputs().get()
        outputs["flame_results"] = self.flame_results
        outputs["zstat_files"] = self.zstat_files
        return outputs


class FFXSummaryInput(BaseInterfaceInputSpec):

    ss_files = InputMultiPath(File(exists=True))
    means = InputMultiPath(File(exists=True))
    timeseries = InputMultiPath(File(exists=True))


class FFXSummaryOutput(TraitedSpec):

    r2_files = OutputMultiPath(File(exists=True))
    tsnr_file = File(exists=True)


class FFXSummary(BaseInterface):

    input_spec = FFXSummaryInput
    output_spec = FFXSummaryOutput

    def _run_interface(self, runtime):

        # Get basic info about the image
        img = nib.load(self.inputs.means[0])
        self.affine = img.get_affine()
        self.header = img.get_header()

        # Save the fixed effects r squared maps
        self.compute_rsquared()

        # Save the tsnr map
        self.compute_tsnr()

        return runtime

    def compute_rsquared(self):
        """Compute and save the full and main rsquared for fixed effects."""
        # First read the total sum of squares for each run
        total_sumsquares = self.sum_squares("sstot")

        for part in ["full", "main"]:

            # Read in the residual sum of squares and take grand sum
            res_sumsquares = self.sum_squares("ssres", part)

            # Calculate the full model R2
            r2 = 1 - res_sumsquares / total_sumsquares

            # Save an image with these data
            self.save_image(r2, "r2_" + part)

    def sum_squares(self, kind, part=None):
        """Sum the sum of squares images across runs."""
        ss_files = self.inputs.ss_files
        stem = kind
        if kind == "ssres":
            stem += "_" + part
        files = [f for f in ss_files if stem in f]
        data = [nib.load(f).get_data() for f in files]
        sums = np.sum(data, axis=0)
        return sums

    def compute_tsnr(self):
        """Compute the toal signal to noise."""
        # Load the mean images from each run and take the grand mean."""
        mean_imgs = [nib.load(f) for f in self.inputs.means]
        mean_data = np.mean([img.get_data() for img in mean_imgs], axis=0)

        # Use the timeseries to get the total number of frames
        ts_lengths = [nib.load(f).shape[-1] for f in self.inputs.timeseries]
        n = sum(ts_lengths)

        # Compute and save tsnr
        res_sumsquares = self.sum_squares("ssres", "full")
        res_std = np.sqrt(res_sumsquares / n)
        tsnr = mean_data / res_std
        tsnr = np.nan_to_num(tsnr)
        self.save_image(tsnr, "tsnr")

    def save_image(self, data, fname):
        """Save data to the output structure."""
        img = nib.Nifti1Image(data, self.affine, self.header)
        img.to_filename(fname + ".nii.gz")

    def _list_outputs(self):

        outputs = self._outputs().get()
        outputs["r2_files"] = [op.abspath("r2_full.nii.gz"),
                               op.abspath("r2_main.nii.gz")]
        outputs["tsnr_file"] = op.abspath("tsnr.nii.gz")
        return outputs


class FFXReportInput(BaseInterfaceInputSpec):

    space = traits.Either("mni", "epi")
    anatomy = File(exists=True)
    tsnr_file = File(exists=True)
    zstat_files = InputMultiPath(File(exists=True))
    r2_files = InputMultiPath(File(exists=True))
    masks = InputMultiPath(File(exists=True))


class FFXReportOutput(TraitedSpec):

    summary_files = OutputMultiPath(File(exists=True))
    zstat_files = OutputMultiPath(Directory(exists=True))


class FFXReport(BaseInterface):

    input_spec = FFXReportInput
    output_spec = FFXReportOutput

    def _run_interface(self, runtime):

        self.step = 1 if self.inputs.space == "epi" else 2
        self.summary_files = []
        self.zstat_files = []
        self.write_mask_image()
        self.write_tsnr_image()
        self.write_r2_images()
        self.write_zstat_images()

        return runtime

    def write_mask_image(self):
        """Show the overlap of each run's mask."""
        mask_imgs = [nib.load(f) for f in self.inputs.masks]
        mask_data = [img.get_data()[..., np.newaxis] for img in mask_imgs]
        each_mask = np.concatenate(mask_data, axis=-1)
        any_mask = each_mask.max(axis=-1)
        self.every_mask = each_mask.min(axis=-1)

        m = Mosaic(self.inputs.anatomy, each_mask, any_mask,
                   stat_interp="nearest", step=self.step, show_mask=False)
        m.plot_mask_edges()
        out_fname = op.abspath("mask_overlap.png")
        self.summary_files.append(out_fname)
        m.savefig(out_fname)
        m.close()

    def write_tsnr_image(self):
        """Show the total subject SNR."""
        m = Mosaic(self.inputs.anatomy,
                   self.inputs.tsnr_file,
                   self.every_mask,
                   step=self.step)
        m.plot_overlay("cube:1.9:.5", 0, alpha=1, fmt="%d")
        out_fname = op.abspath("tsnr.png")
        self.summary_files.append(out_fname)
        m.savefig(out_fname)
        m.close()

    def write_r2_images(self):
        """Show the fixed effects model fit."""
        cmaps = ["cube:2:0", "cube:2.6:0"]
        for fname, cmap in zip(self.inputs.r2_files, cmaps):
            m = Mosaic(self.inputs.anatomy, fname,
                       self.every_mask, step=self.step)
            m.plot_overlay(cmap, 0, alpha=.85)
            out_fname = op.abspath(nii_to_png(fname))
            self.summary_files.append(out_fname)
            m.savefig(out_fname)
            m.close()

    def write_zstat_images(self):
        """Show the fixed effects inference maps."""
        for fname in self.inputs.zstat_files:
            m = Mosaic(self.inputs.anatomy, fname,
                       self.every_mask, step=self.step)
            m.plot_activation(pos_cmap="Reds_r", neg_cmap="Blues",
                              thresh=2.3, alpha=.85)
            contrast = fname.split("/")[-2]
            os.mkdir(contrast)
            out_fname = op.join(contrast, "zstat1.png")
            self.zstat_files.append(op.abspath(contrast))
            m.savefig(out_fname)
            m.close()

    def _list_outputs(self):

        outputs = self._outputs().get()
        outputs["summary_files"] = self.summary_files
        outputs["zstat_files"] = self.zstat_files
        return outputs
